@@ -1,8 +1,8 @@
 #!/usr/bin/perl -w
 
-# $Id: oinkmaster.pl,v 1.359 2004/10/01 18:30:26 andreas_o Exp $ #
+# $Id: oinkmaster.pl,v 1.378 2005/04/12 17:50:36 andreas_o Exp $ #
 
-# Copyright (c) 2001-2004 Andreas stling <andreaso@it.su.se>
+# Copyright (c) 2001-2005 Andreas stling <andreaso@it.su.se>
 # All rights reserved.
 #
 #  Redistribution and use in source and binary forms, with or
@@ -53,9 +53,10 @@ sub parse_cmdline($);
 sub read_config($ $);
 sub sanity_check();
 sub download_file($ $);
-sub unpack_rules_archive($ $);
+sub unpack_rules_archive($ $ $);
+sub join_tmp_rules_dirs($ $ @);
 sub process_rules($ $ $ $ $ $);
-sub process_rule($ $ $ $ $ $ $);
+sub process_rule($ $ $ $ $ $ $ $);
 sub setup_rules_hash($ $);
 sub get_first_only($ $ $);
 sub print_changes($ $);
@@ -63,7 +64,6 @@ sub print_changetype($ $ $ $);
 sub print_summary_change($ $);
 sub make_backup($ $);
 sub get_changes($ $ $);
-sub get_new_filenames($ $);
 sub update_rules($ @);
 sub is_in_path($);
 sub get_next_entry($ $ $ $ $ $);
@@ -80,10 +80,10 @@ sub catch_sigint();
 sub clean_exit($);
 
 
-my $VERSION            = 'Oinkmaster v1.1 by Andreas stling <andreaso@it.su.se>';
+my $VERSION            = 'Oinkmaster v1.2, Copyright (C) 2001-2005 Andreas stling <andreaso@it.su.se>';
 my $OUTFILE            = 'snortrules.tar.gz';
 my $RULES_DIR          = 'rules';
-my $DIST_SNORT_CONF    = "$RULES_DIR/snort.conf";
+my $DIST_SNORT_CONF    = 'snort.conf';
 
 my $PRINT_NEW          = 1;
 my $PRINT_OLD          = 2;
@@ -214,20 +214,24 @@ if ($config{config_test_mode}) {
 
 umask($config{umask}) if exists($config{umask});
 
-# Download the rules archive. Will exit if it fails.
-download_file("$config{url}", "$tmpdir/$OUTFILE");
+# Download and unpack all the rules archives into separate tmp dirs.
+my @url_tmpdirs;
+foreach my $url (@{$config{url}}) {
+    my $url_tmpdir = tempdir("url.XXXXXXXXXX", DIR => $tmpdir)
+      or clean_exit("could not create temporary directory in $tmpdir: $!");
+    push(@url_tmpdirs, "$url_tmpdir/$RULES_DIR");
+    download_file($url, "$url_tmpdir/$OUTFILE");
+    unpack_rules_archive("$url", "$url_tmpdir/$OUTFILE", $RULES_DIR);
+}
 
-# Verify and unpack archive. This will leave us with a directory
-# called $RULES_DIR in the same directory as the archive, containing the
-# new rules. Will exit if something fails.
-unpack_rules_archive("$tmpdir/$OUTFILE", $RULES_DIR);
-
-# Create list of new files that we care about from the downloaded
+# Copy all rules files from the tmp dirs into $RULES_DIR in the tmp directory.
+# File matching 'skipfile' a directive will not be copied.
 # Filenames (with full path) will be stored as %new_files{filename}.
-my $num_files = get_new_filenames(\my %new_files, "$tmpdir/$RULES_DIR");
+# Will exit in case of duplicate filenames.
+my $num_files = join_tmp_rules_dirs("$tmpdir/$RULES_DIR", \my %new_files, @url_tmpdirs);
 
 # Make sure we have at least the minimum number of files.
-clean_exit("not enough rules files in downloaded archive (is it broken?)\n".
+clean_exit("not enough rules files in downloaded rules archive(s).\n".
            "Number of rules files is $num_files but minimum is set to $config{min_files}.")
   if ($num_files < $config{min_files});
 
@@ -235,7 +239,7 @@ clean_exit("not enough rules files in downloaded archive (is it broken?)\n".
 my %rh_tmp = setup_rules_hash(\%new_files, $config{output_dir});
 
 # Disable/modify/clean downloaded rules.
-my $num_rules = process_rules(\%{$config{sid_modify_list}},
+my $num_rules = process_rules(\@{$config{sid_modify_list}},
                               \%{$config{sid_disable_list}},
                               \%{$config{sid_enable_list}},
                               \%{$config{sid_local_list}},
@@ -243,7 +247,7 @@ my $num_rules = process_rules(\%{$config{sid_modify_list}},
                               \%new_files);
 
 # Make sure we have at least the minimum number of rules.
-clean_exit("not enough rules in downloaded archive (is it broken?)\n".
+clean_exit("not enough rules in downloaded archive(s).\n".
            "Number of rules is $num_rules but minimum is set to $config{min_rules}.")
   if ($num_rules < $config{min_rules});
 
@@ -253,8 +257,8 @@ my %rh = setup_rules_hash(\%new_files, $config{output_dir});
 # Compare the new rules to the old ones.
 my %changes = get_changes(\%rh, \%new_files, $RULES_DIR);
 
-# Check for variables that exist in dist snort.conf but not in local snort.conf.
-get_new_vars(\%changes, $config{varfile}, "$tmpdir/$DIST_SNORT_CONF")
+# Check for variables that exist in dist snort.conf(s) but not in local snort.conf.
+get_new_vars(\%changes, $config{varfile}, \@url_tmpdirs)
   if ($config{update_vars});
 
 
@@ -341,10 +345,11 @@ Options:
 -r        Check for rules files that exist in the output directory
           but not in the downloaded rules archive
 -s        Leave out details in rules results, just print SID, msg and filename
--T        Test configuration and then exit
--u <url>  Download from this URL instead of the one in the configuration file
-          (must be http://, https://, ftp://, file:// or scp:// ... .tar.gz)
--U <file> Merge new variables from downloaded snort.conf into <file>
+-T        Config test - just check configuration file(s) for errors/warnings
+-u <url>  Download from this URL instead of URL(s) in the configuration file
+          (http://, https://, ftp://, file:// or scp:// ... .tar.gz|.gz)
+          May be specified multiple times to grab multiple rules archives
+-U <file> Merge new variables from downloaded snort.conf(s) into <file>
 -v        Verbose mode (debug)
 -V        Show version and exit
 
@@ -375,7 +380,7 @@ sub parse_cmdline($)
         "r"   => \$$cfg_ref{check_removed},
         "s"   => \$$cfg_ref{summary_output},
         "T"   => \$$cfg_ref{config_test_mode},
-        "u=s" => \$$cfg_ref{url},
+        "u=s" => \@{$$cfg_ref{url}},
         "U=s" => \$$cfg_ref{varfile},
         "v"   => \$$cfg_ref{verbose},
         "V"   => sub {
@@ -403,7 +408,7 @@ sub parse_cmdline($)
     }
 
   # Mark that url was set on command line (so we don't override it later).
-    $$cfg_ref{cmdline_url} = 1 if ($$cfg_ref{url});
+    $$cfg_ref{cmdline_url} = 1 if ($#{$config{url}} > -1);
 }
 
 
@@ -476,7 +481,7 @@ sub read_config($ $)
         next unless (/\S/);
 
       # Use a template and make $_ a "modifysid" line.
-        if (/^use_template\s+(\S+)\s+([\d\s,\*]+)\s*(".*")*(?:#.*)*/i) {
+        if (/^use_template\s+(\S+)\s+(\S+[^"]*)\s*(".*")*(?:#.*)*/i) {
             my ($template_name, $sid, $args) = ($1, $2, $3);
 
             if (exists($templates{$template_name})) {
@@ -520,11 +525,11 @@ sub read_config($ $)
                 $templates{$template_name} = $template;
             }
 
-      # modifysid <SID[,SID, ...]> "substthis" | "withthis"
-        } elsif (/^modifysids*\s+(\d+.*|\*)\s+"(.+)"\s+\|\s+"(.*)"\s*(?:#.*)*$/i) {
+      # modifysid <SIDORFILE[,SIDORFILE, ...]> "substthis" | "withthis"
+        } elsif (/^modifysids*\s+(\S+.*)\s+"(.+)"\s+\|\s+"(.*)"\s*(?:#.*)*$/i) {
             my ($sid_list, $subst, $repl) = ($1, $2, $3);
             warn("WARNING: line $linenum in $config_file is invalid, ignoring\n")
-              unless(parse_mod_expr(\%{$$cfg_ref{sid_modify_list}},
+              unless(parse_mod_expr(\@{$$cfg_ref{sid_modify_list}},
                                     $sid_list, $subst, $repl));
 
       # disablesid <SID[,SID, ...]>
@@ -573,7 +578,7 @@ sub read_config($ $)
 	    }
 
 	} elsif (/^url\s*=\s*(.*)/i) {
-	    $$cfg_ref{url} = $1
+            push(@{$$cfg_ref{url}}, $1)
               unless ($$cfg_ref{cmdline_url});
 
 	} elsif (/^path\s*=\s*(.+)/i) {
@@ -693,9 +698,18 @@ sub sanity_check()
     }
 
   # Make sure $url is defined (either by -u <url> or url=... in the conf).
-    clean_exit("incorrect URL or URL not specified in either configuration file or command line.")
-      unless (defined($config{'url'}) &&
-        (($config{'url'}) = $config{'url'} =~ /^((?:https*|ftp|file|scp):\/\/.+\.tar\.gz)$/));
+    clean_exit("URL not specified. Specify at least one \"url=<url>\" in the \n".
+               "Oinkmaster configuration file or use the \"-u <url>\" argument")
+      if ($#{$config{url}} == -1);
+
+  # Make sure all urls look ok, and untaint them.
+    my @urls = @{$config{url}};
+    $#{$config{url}} = -1;
+    foreach my $url (@urls) {
+        clean_exit("incorrect URL: \"$url\"")
+          unless ($url =~ /^((?:https*|ftp|file|scp):\/\/.+\.(?:tar\.gz|tgz))$/);
+        push(@{$config{url}}, $1);
+    }
 
   # Wget must be found if url is http[s]:// or ftp://.
     if ($config{use_external_bins}) {
@@ -735,15 +749,19 @@ sub sanity_check()
         next if (/^\.\.?$/ || exists($config{file_ignore_list}{$_}));
 
         if (/$config{update_files}/) {
-            clean_exit("no read permission on \"$config{output_dir}/$_\"\n".
-                       "(Read permission is required on all rules files ".
-                       "inside the output directory.)\n")
-              unless (-r "$config{output_dir}/$_");
+            unless (-r "$config{output_dir}/$_") {
+                closedir(OUTDIR);
+                clean_exit("no read permission on \"$config{output_dir}/$_\"\n".
+                           "(Read permission is required on all rules files ".
+                           "inside the output directory.)\n")
+            }
 
-            clean_exit("no write permission on \"$config{output_dir}/$_\"\n".
-                       "(Write permission is required on all rules files ".
-                       "inside the output directory.)\n")
-              if (!$config{careful} && !-w "$config{output_dir}/$_");
+            if (!$config{careful} && !-w "$config{output_dir}/$_") {
+                closedir(OUTDIR);
+                clean_exit("no write permission on \"$config{output_dir}/$_\"\n".
+                           "(Write permission is required on all rules files ".
+                           "inside the output directory.)\n")
+            }
 	}
     }
 
@@ -799,7 +817,7 @@ sub download_file($ $)
 
         if ($config{verbose}) {
             print STDERR "\n";
-            clean_exit("could not download file")
+            clean_exit("could not download from $obfuscated_url")
               if (system("wget", "-v", "-O", "$localfile", "$url"));
         } else {
             if (system("wget", "-v", "-o", "$log", "-O", "$localfile", "$url")) {
@@ -807,7 +825,7 @@ sub download_file($ $)
                   or clean_exit("could not open $log for reading: $!");
                 my @log = <LOG>;
                 close(LOG);
-                clean_exit("could not download file. Output from wget follows:\n\n @log");
+                clean_exit("could not download from $obfuscated_url. Output from wget follows:\n\n @log");
             }
             print STDERR "done.\n" unless ($config{quiet});
         }
@@ -822,7 +840,7 @@ sub download_file($ $)
 	my $request = HTTP::Request->new(GET => $url);
 	my $response = $ua->request($request, $localfile);
 
-        clean_exit("could not download file: " . $response->status_line)
+        clean_exit("could not download from $obfuscated_url: " . $response->status_line)
           unless $response->is_success;
 
         print "done.\n" unless ($config{quiet});
@@ -869,22 +887,74 @@ sub download_file($ $)
     }
 
   # Make sure the downloaded file actually exists.
-    clean_exit("failed to download file: ".
+    clean_exit("failed to download $url: ".
                "local target file $localfile doesn't exist after download.")
       unless (-e "$localfile");
 
   # Also make sure it's at least non-empty.
-    clean_exit("failed to download file: local target file $localfile is empty ".
+    clean_exit("failed to download $url: local target file $localfile is empty ".
                "after download (perhaps you're out of diskspace or file in url is empty?)")
       unless (-s "$localfile");
 }
 
 
 
+# Copy all rules files from the tmp dirs (one for each url)
+# into a single directory inside the tmp dir, except for files
+# matching a 'skipfile' directive'.
+# Will exit in case of colliding filenames.
+sub join_tmp_rules_dirs($ $ @)
+{
+    my $rules_dir     = shift;
+    my $new_files_ref = shift;
+    my @url_tmpdirs   = @_;
+
+    my %rules_files;
+
+    clean_exit("failed to create directory \"$rules_dir\": $!")
+      unless (mkdir($rules_dir));
+
+    foreach my $url_tmpdir (@url_tmpdirs) {
+        opendir(URL_TMPDIR, "$url_tmpdir")
+          or clean_exit("could not open directory \"$url_tmpdir\": $!");
+
+        while ($_ = readdir(URL_TMPDIR)) {
+            next if (/^\.\.?$/ || exists($config{file_ignore_list}{$_}) || !/$config{update_files}/);
+
+            if (exists($rules_files{$_})) {
+                closedir(URL_TMPDIR);
+                clean_exit("a file called \"$_\" exists in multiple rules archives")
+            }
+
+          # Make sure it's a regular file.
+            unless (-f "$url_tmpdir/$_" && !-l "$url_tmpdir/$_") {
+                closedir(URL_TMPDIR);
+                clean_exit("downloaded \"$_\" is not a regular file.")
+            }
+
+            $rules_files{$_} = 1;
+            $$new_files_ref{"$rules_dir/$_"} = 1;
+
+            my $src_file = untaint_path("$url_tmpdir/$_");
+            unless (copy("$src_file", "$rules_dir")) {
+                closedir(URL_TMPDIR);
+                clean_exit("could not copy \"$src_file\" to \"$rules_dir\": $!");
+            }
+        }
+
+        closedir(URL_TMPDIR);
+    }
+
+    return (keys(%$new_files_ref));
+}
+
+
+
 # Make a few basic sanity checks on the rules archive and then
 # uncompress/untar it if everything looked ok.
-sub unpack_rules_archive($ $)
+sub unpack_rules_archive($ $ $)
 {
+    my $url       = shift;  # only used when printing warnings/errors
     my $archive   = shift;
     my $rules_dir = shift;
 
@@ -893,26 +963,26 @@ sub unpack_rules_archive($ $)
     my $old_dir = untaint_path(File::Spec->rel2abs(File::Spec->curdir()));
 
     my $dir = dirname($archive);
-    chdir("$dir") or clean_exit("could not change directory to \"$dir\": $!");
+    chdir("$dir") or clean_exit("$url: could not change directory to \"$dir\": $!");
 
     if ($config{use_external_bins}) {
 
       # Run integrity check on the gzip file.
-        clean_exit("integrity check on gzip file failed (file transfer failed or ".
+        clean_exit("$url: integrity check on gzip file failed (file transfer failed or ".
                    "file in URL not in gzip format?).")
           if (system("gzip", "-t", "$archive"));
 
       # Decompress it.
         system("gzip", "-d", "$archive")
-          and clean_exit("unable to uncompress $archive.");
+          and clean_exit("$url: unable to uncompress $archive.");
 
-      # Suffix has now changed from .tar.gz to .tar.
+      # Suffix has now changed from .tar.gz|.tgz to .tar.
         $archive =~ s/\.gz$//;
 
       # Make sure the .tar file now exists.
       # (Gzip may not return an error if it was not a gzipped file...)
-        clean_exit("failed to unpack gzip file (file transfer failed or ".
-                   "file in URL not in gzip format?).")
+        clean_exit("$url: failed to unpack gzip file (file transfer failed or ".
+                   "file in URL not in tar'ed gzip format?).")
           unless (-e  "$archive");
 
         my $stdout_file = "$tmpdir/tar_content.out";
@@ -926,7 +996,7 @@ sub unpack_rules_archive($ $)
         open(STDOUT, ">&OLDOUT") or clean_exit("could not dup STDOUT: $!");
         close(OLDOUT);
 
-        clean_exit("could not list files in tar archive (is it broken?)")
+        clean_exit("$url: could not list files in tar archive (is it broken?)")
           if ($ret);
 
         open(TAR, "$stdout_file") or clean_exit("failed to open $stdout_file: $!");
@@ -936,13 +1006,13 @@ sub unpack_rules_archive($ $)
  # use_external_bins=0
     } else {
         $tar = Archive::Tar->new($archive, 1);
-        clean_exit("could not read $archive\n")
+        clean_exit("$url: could not read $archive\n")
           unless (defined($tar));
         @tar_content = $tar->list_files();
     }
 
   # Make sure we could grab some content from the tarball.
-    clean_exit("could not list files in tar archive (is it broken?)")
+    clean_exit("$url: could not list files in tar archive (is it broken?)")
       if ($#tar_content < 0);
 
   # For each filename in the archive, do some basic sanity checks.
@@ -950,17 +1020,17 @@ sub unpack_rules_archive($ $)
        chomp($filename);
 
       # We don't want absolute filename.
-        clean_exit("rules archive contains absolute filename. ".
+        clean_exit("$url: rules archive contains absolute filename. ".
                    "Offending file/line:\n$filename")
           if ($filename =~ /^\//);
 
       # We don't want to have any weird characters anywhere in the filename.
-        clean_exit("illegal character in filename in tar archive. Allowed are ".
+        clean_exit("$url: illegal character in filename in tar archive. Allowed are ".
                    "$OK_PATH_CHARS\nOffending file/line:\n$filename")
           if ($config{use_path_checks} && $filename =~ /[^$OK_PATH_CHARS]/);
 
       # We don't want to unpack any "../../" junk (check is useless now though).
-        clean_exit("filename in tar archive contains \"..\".\n".
+        clean_exit("$url: filename in tar archive contains \"..\".\n".
                    "Offending file/line:\n$filename")
           if ($filename =~ /\.\./);
     }
@@ -986,9 +1056,25 @@ sub unpack_rules_archive($ $)
         }
     }
 
-    clean_exit("no \"$rules_dir\" directory found in tar file.")
+  # Make sure that non-empty rules directory existed in archive.
+  # We permit empty rules directory if min_files is set to 0 though.
+    clean_exit("$url: no \"$rules_dir\" directory found in tar file.")
       unless (-d "$dir/$rules_dir");
 
+    my $num_files = 0;
+    opendir(RULESDIR, "$dir/$rules_dir")
+      or clean_exit("could not open directory \"$dir/$rules_dir\": $!");
+
+    while ($_ = readdir(RULESDIR)) {
+        next if (/^\.\.?$/);
+        $num_files++;
+    }
+
+    closedir(RULESDIR);
+
+    clean_exit("$url: directory \"$rules_dir\" in unpacked archive is empty")
+      if ($num_files == 0 && $config{min_files} != 0);
+
     chdir($old_dir)
       or clean_exit("could not change directory back to $old_dir: $!");
 
@@ -1015,6 +1101,7 @@ sub process_rules($ $ $ $ $ $)
         disabled => 0,
         enabled  => 0,
         modified => 0,
+        total    => 0,
     );
 
     warn("WARNING: all rules that are disabled by default will be enabled\n")
@@ -1050,10 +1137,14 @@ sub process_rules($ $ $ $ $ $)
                 multi  => $multi,
             );
 
-          # modify/disable/enable this rule as requested. Possible verbose 
-          # messages and warnings will be printed.
-            process_rule($modify_sid_ref, $disable_sid_ref, $enable_sid_ref, 
-                         \%rule, $sid, \%stats, 1);
+          # modify/disable/enable this rule as requested unless there is a matching
+          # localsid statement. Possible verbose messages and warnings will be printed.
+            unless (exists($$local_sid_ref{$sid})) {
+                process_rule($modify_sid_ref, $disable_sid_ref, $enable_sid_ref,
+                             \%rule, $sid, \%stats, 1, basename($file));
+            }
+
+            $stats{total}++;
 
             $single = $rule{single};
             $multi  = $rule{multi};
@@ -1065,13 +1156,13 @@ sub process_rules($ $ $ $ $ $)
 
           # Is it a dup? If so, see if this seems to be more recent (higher rev).
             if (exists($sids{$sid})) {
-                warn("WARNING: duplicate SID in downloaded archive, SID=$sid\n")
+                warn("\nWARNING: duplicate SID in downloaded archive, SID=$sid\n")
                   unless($config{super_quiet});
 
                 my ($old_rev) = ($sids{$sid}{single} =~ /\brev\s*:\s*(\d+)\s*;/);
                 my ($new_rev) = ($single             =~ /\brev\s*:\s*(\d+)\s*;/);
 
-              # This is so rules with a rev gets higher prio than 
+              # This is so rules with a rev gets higher prio than
               # rules without any rev.
                 $old_rev = -1 unless (defined($old_rev));
                 $new_rev = -1 unless (defined($new_rev));
@@ -1095,7 +1186,6 @@ sub process_rules($ $ $ $ $ $)
 
   # Phase #2 - read all rules files again, but when writing active rules
   # back to the files, use the one stored in the sid hash (which is free of dups).
-
     foreach my $file (sort(keys(%$newfiles_ref))) {
 
         open(INFILE, "<", "$file")
@@ -1155,7 +1245,7 @@ sub process_rules($ $ $ $ $ $)
           # We send the stats to a dummy var as this was collected on the
           # first phase as well.
             process_rule($modify_sid_ref, $disable_sid_ref, $enable_sid_ref,
-                         \%rule, $sid, \my %unused_stats, 0);
+                         \%rule, $sid, \my %unused_stats, 0, basename($file));
 
             $single = $rule{single};
             $multi  = $rule{multi};
@@ -1186,15 +1276,9 @@ sub process_rules($ $ $ $ $ $)
                  "modified $stats{modified}, total=$stats{total}\n"
       unless ($config{quiet});
 
-  # Warn on attempt at processing non-existent sids, unless quiet mode.
+  # Print warnings on attempt at enablesid/disablesid/localsid on non-existent
+  # rule if we're in verbose mode.
     if ($config{verbose}) {
-
-        foreach my $sid (keys(%$modify_sid_ref)) {
-            next unless ($sid =~ /^\d+$/);    # don't warn on wildcard match
-            warn("WARNING: attempt to use \"modifysid\" on non-existent SID $sid\n")
-              unless (exists($sids{$sid}));
-        }
-
         foreach my $sid (keys(%$enable_sid_ref)) {
             warn("WARNING: attempt to use \"enablesid\" on non-existent SID $sid\n")
               unless (exists($sids{$sid}));
@@ -1211,6 +1295,31 @@ sub process_rules($ $ $ $ $ $)
         }
     }
 
+  # Print warnings on attempt at modifysid'ing non-existent stuff in all modes
+  # except super quiet, as they are usually more important.
+    unless ($config{super_quiet}) {
+        my %new_files;
+        foreach my $file (sort(keys(%$newfiles_ref))) {
+            $new_files{basename($file)} = 1;
+        }
+
+        my %mod_tmp;
+        foreach my $mod_expr (@$modify_sid_ref) {
+            my ($type, $arg) = ($mod_expr->[2], $mod_expr->[3]);
+            $mod_tmp{$type}{$arg} = 1;
+        }
+
+        foreach my $sid (keys(%{$mod_tmp{sid}})) {
+            warn("WARNING: attempt to use \"modifysid\" on non-existent SID $sid\n")
+              unless (exists($sids{$sid}));
+        }
+
+        foreach my $file (keys(%{$mod_tmp{file}})) {
+            warn("WARNING: attempt to use \"modifysid\" on non-existent file $file\n")
+              unless(exists($new_files{$file}));
+        }
+    }
+
   # Return total number of valid rules.
     return ($stats{total});
 }
@@ -1218,7 +1327,7 @@ sub process_rules($ $ $ $ $ $)
 
 
 # Process (modify/enable/disable) a rule as requested.
-sub process_rule($ $ $ $ $ $ $)
+sub process_rule($ $ $ $ $ $ $ $)
 {
     my $modify_sid_ref  = shift;
     my $disable_sid_ref = shift;
@@ -1227,13 +1336,12 @@ sub process_rule($ $ $ $ $ $ $)
     my $sid             = shift;
     my $stats_ref       = shift;
     my $print_messages  = shift;
+    my $filename        = shift;
 
   # Just for easier access.
     my $single = $$rule_ref{single};
     my $multi  = $$rule_ref{multi};
 
-    $$stats_ref{total}++;
-
   # Some rules may be commented out by default.
   # Enable them if -e is specified (both single-line and multi-line,
   # version, because we don't know which version one we're going to
@@ -1251,33 +1359,34 @@ sub process_rule($ $ $ $ $ $ $)
   # modifysid that's no good since we don't know where in the
   # rule the trailing backslashes and newlines are going to be
   # and we don't want them to affect the regexp.
-    my @all_mod = @{$$modify_sid_ref{'*'}}
-      if (exists($$modify_sid_ref{'*'}));
+    foreach my $mod_expr (@$modify_sid_ref) {
+        my ($subst, $repl, $type, $arg) =
+          ($mod_expr->[0], $mod_expr->[1], $mod_expr->[2], $mod_expr->[3]);
 
-    my @sid_mod = @{$$modify_sid_ref{$sid}}
-      if (exists($$modify_sid_ref{$sid}));
+        if ($type eq "wildcard" || ($type eq "sid" && $sid eq $arg) ||
+          ($type eq "file" && $filename eq $arg)) {
 
-    foreach my $mod_expr (@sid_mod, @all_mod) {
+            if ($single =~ /$subst/si) {
+                print STDERR "Modifying rule, SID=$sid, filename=$filename, ".
+                             "match type=$type, subst=$subst, ".
+                             "repl=$repl\nBefore: $single"
+                  if ($print_messages && $config{verbose});
 
-        my ($subst, $repl) = ($mod_expr->[0], $mod_expr->[1]);
-        if ($single =~ /$subst/si) {
-            print STDERR "Modifying SID $sid, subst=$subst, ".
-                         "repl=$repl\nBefore: $single\n"
-	      if ($print_messages && $config{verbose});
+              # Do the substitution on the single-line version and put it
+              # back in $multi.
+                $single =~ s/$subst/$repl/eei;
+                $multi = $single;
 
-          # Do the substitution on the single-line version and put it
-          # back in $multi.
-            $single =~ s/$subst/$repl/eei;
-            $multi = $single;
+      	        print STDERR "After:  $single\n"
+                  if ($print_messages && $config{verbose});
 
-  	    print STDERR "After:  $single\n"
-              if ($print_messages && $config{verbose});
-
-            $$stats_ref{modified}++;
-        } else {
-            print STDERR "WARNING: SID $sid does not match ".
-                         "modifysid expression \"$subst\", skipping\n"
-              if ($print_messages && !exists($$modify_sid_ref{'*'}));
+                $$stats_ref{modified}++;
+            } else {
+                if (!$config{super_quiet} && $print_messages && $type eq "sid") {
+                    print STDERR "WARNING: SID $sid does not match ".
+                                 "modifysid expression \"$subst\", skipping.\n";
+                }
+            }
         }
     }
 
@@ -1319,10 +1428,6 @@ sub setup_rules_hash($ $)
         warn("\nWARNING: downloaded rules file $file is empty\n")
           if (!-s "$file" && $config{verbose});
 
-      # Make sure it's a regular file.
-        clean_exit("\"$file\" is not a regular file.")
-          unless (-f "$file" && !-l "$file");
-
         open(NEWFILE, "<", "$file")
           or clean_exit("could not open $file for reading: $!");
         my @newfile = <NEWFILE>;
@@ -1335,7 +1440,7 @@ sub setup_rules_hash($ $)
 
 	while (get_next_entry(\@newfile, \$single, \$multi, \$nonrule, \$msg, \$sid)) {
 	    if (defined($single)) {
-		$rh{new}{rules}{"$file"}{"$sid"} = $single;
+  	        $rh{new}{rules}{"$file"}{"$sid"} = $single;
 	    } else {
 	        push(@{$rh{new}{other}{"$file"}}, $nonrule);
 	    }
@@ -1351,8 +1456,8 @@ sub setup_rules_hash($ $)
 
 	    while (get_next_entry(\@oldfile, \$single, \$multi, \$nonrule, undef, \$sid)) {
 	        if (defined($single)) {
-		    warn("WARNING: duplicate SID in your local rules, SID ".
-                         "$sid exists multiple times, please fix this manually!\n")
+		    warn("\nWARNING: duplicate SID in your local rules, SID ".
+                         "$sid exists multiple times, you may need to fix this manually!\n")
 		      if (exists($old_sids{$sid}));
 
 	  	    $rh{old}{rules}{"$file"}{"$sid"} = $single;
@@ -1496,8 +1601,12 @@ sub print_changes($ $)
     my $ch_ref = shift;
     my $rh_ref = shift;
 
-    print "\n[***] Results from Oinkmaster started " .
-          scalar(localtime) . " [***]\n";
+    my ($sec, $min, $hour, $mday, $mon, $year) = (localtime)[0 .. 5];
+
+    my $date = sprintf("%4d%02d%02d %02d:%02d:%02d",
+                       $year + 1900, $mon + 1, $mday, $hour, $min, $sec);
+
+    print "\n[***] Results from Oinkmaster started $date [***]\n";
 
   # Print new variables.
     if ($config{update_vars}) {
@@ -1524,7 +1633,7 @@ sub print_changes($ $)
             print_summary_change(\%{$$ch_ref{rules}{added}}, $rh_ref);
         } else {
             print_changetype($PRINT_NEW, "Added to",
-                                 \%{$$ch_ref{rules}{added}}, $rh_ref);
+                             \%{$$ch_ref{rules}{added}}, $rh_ref);
         }
     }
 
@@ -1856,30 +1965,6 @@ sub get_changes($ $ $)
 
 
 
-# Create list of new files (with full path) that we care about.
-# I.e. files that match the 'update_files' regexp and isn't listed
-# in the ignore list.
-sub get_new_filenames($ $)
-{
-    my $new_files_ref = shift;
-    my $new_rules_dir = shift;
-
-    opendir(NEWRULES, "$new_rules_dir")
-      or clean_exit("could not open directory $new_rules_dir: $!");
-
-    while ($_ = readdir(NEWRULES)) {
-        next if (/^\.\.?$/);
-        $new_files{"$new_rules_dir/$_"}++
-          if (/$config{update_files}/ && !exists($config{file_ignore_list}{$_}));
-    }
-    closedir(NEWRULES);
-
-  # Return number of new interesting filenames.
-    return (keys(%$new_files_ref));
-}
-
-
-
 # Simply copy the modified rules files to the output directory.
 sub update_rules($ @)
 {
@@ -1948,6 +2033,9 @@ sub get_next_entry($ $ $ $ $ $)
     my $disabled = 0;
     my $broken   = 0;
 
+    chomp($line);
+    $line .= "\n";
+
   # Possible beginning of multi-line rule?
     if ($line =~ /$MULTILINE_RULE_REGEXP/oi) {
         $$single_ref = $line;
@@ -2066,22 +2154,35 @@ sub get_next_entry($ $ $ $ $ $)
 # Look for variables that exist in dist snort.conf but not in local snort.conf.
 sub get_new_vars($ $ $)
 {
-    my $ch_ref     = shift;
-    my $local_conf = shift;
-    my $dist_conf  = shift;
+    my $ch_ref          = shift;
+    my $local_conf      = shift;
+    my $url_tmpdirs_ref = shift;
+
     my @new_vars;
     my %old_vars;
 
-    unless (-e "$dist_conf") {
-        $_ = basename($dist_conf);
-        warn("WARNING: no $_ found in downloaded archive, ".
-             "aborting check for new variables\n");
-        return;
+    my $confs_found = 0;
+
+    foreach my $dir (@$url_tmpdirs_ref) {
+        if (-e "$dir/$DIST_SNORT_CONF") {
+            $confs_found++;
+        }
     }
 
-    print STDERR "Looking for new variables... "
-      unless ($config{quiet});
+    my $base_conf = basename($DIST_SNORT_CONF);
+    unless ($confs_found) {
+        unless ($config{quiet}) {
+            warn("WARNING: no $base_conf found in downloaded archive(s), ".
+                 "aborting check for new variables\n");
+            return;
+        }
+    }
 
+    unless ($config{quiet}) {
+        print STDERR "Found $confs_found $DIST_SNORT_CONF file";
+        print STDERR "s" if ($confs_found > 1);
+        print STDERR ", checking for new variables... ";
+    }
 
   # Read in variable names from old file.
     open(LOCAL_CONF, "<", "$local_conf")
@@ -2096,17 +2197,31 @@ sub get_new_vars($ $ $)
     close(LOCAL_CONF);
 
 
-  # Read in variables from new file.
-    open(DIST_CONF, "<", "$dist_conf")
-      or clean_exit("could not open $dist_conf for reading: $!");
+  # Read in variables from new files.
+    foreach my $dir (@$url_tmpdirs_ref) {
+        my $conf = "$dir/$DIST_SNORT_CONF";
+        if (-e "$conf") {
 
-    while ($_ = <DIST_CONF>) {
-        push(@new_vars, $_)
-          if (/$VAR_REGEXP/i && !exists($old_vars{lc($1)}));
+            open(DIST_CONF, "<", "$conf")
+              or clean_exit("could not open $conf for reading: $!");
+
+            while ($_ = <DIST_CONF>) {
+                s/^\s*//;
+
+                if (/$VAR_REGEXP/i && !exists($old_vars{lc($1)})) {
+                    if (/\\\n/) {
+                        warn("\nWARNING: can not handle variables split over multiple lines\n")
+                          unless ($config{quiet});
+                    } else {
+                        push(@new_vars, $_)
+                    }
+                }
+            }
+
+            close(DIST_CONF);
+        }
     }
 
-    close(DIST_CONF);
-
     @{$$ch_ref{new_vars}} = @new_vars;
 
     print STDERR "done.\n"
@@ -2178,14 +2293,20 @@ sub msdos_to_cygwin_path($)
 sub parse_mod_expr($ $ $ $)
 {
     my $mod_list_ref = shift;  # where to store valid entries
-    my $sid_list     = shift;  # comma-separated list of SIDs
+    my $sid_arg_list = shift;  # comma-separated list of SIDs/files or wildcard
     my $subst        = shift;  # regexp to look for
     my $repl         = shift;  # regexp to replace it with
 
-    $sid_list =~ s/\s+$//;
+    $sid_arg_list =~ s/\s+$//;
 
-    foreach my $sid (split(/\s*,\s*/, $sid_list)) {
-        return (0) unless ($sid =~ /^\d+$/ || $sid eq "*");
+    foreach my $sid_arg (split(/\s*,\s*/, $sid_arg_list)) {
+        my $type = "";
+
+        $type = "sid"      if ($sid_arg =~ /^\d+$/);
+        $type = "file"     if ($sid_arg =~ /^\S+.*\.\S+$/);
+        $type = "wildcard" if ($sid_arg eq "*");
+
+        return (0) unless ($type);
 
       # Make sure the regexp is valid.
         my $repl_qq = "qq/$repl/";
@@ -2200,7 +2321,7 @@ sub parse_mod_expr($ $ $ $)
             return (0);
         }
 
-        push(@{$$mod_list_ref{$sid}}, [$subst, $repl_qq]);
+        push(@$mod_list_ref, [$subst, $repl_qq, $type, $sid_arg]);
     }
 
     return (1);
